/******************************************************************************
 *
 * ArcBall.cpp - used for 3D rotations
 *
 * Adapted from starter code used for a Princeton CS assignment, see
 * http://www.cs.princeton.edu/courses/archive/fall01/cs426/assignments/assn3/assn3.html
 *
 * In original code, this code falls mostly in arcball.cpp and quaternion.cpp
 *
 * *** See ArcBall.h for more information
 *
 *  02/09/02: D. Hazen--File created 
 *  02/20/02: D. Hazen--Fixed display and rotation problems
 *  03/19/02: D. Hazen--Finished adding comments to the file
 *****************************************************************************/

#include "StdAfx.h"
#include "ArcBall.h"

// unit rotation matrix
static HMatrix mID = {
  { 1, 0, 0, 0 },
  { 0, 1, 0, 0 },
  { 0, 0, 1, 0 },
  { 0, 0, 0, 1 } };

// I have no clue...
static float OTHER_AXIS[][4] = {
  {-0.48f, 0.80f, 0.36f, 1} };

/////////////////////////////////////////////////////////////////////////////
// CQuat Definitions

const CQuat
  CQuat::ZERO,
  CQuat::ONE(0,0,0,1);

CQuat::CQuat(const CQuat & copy) {
    x = copy.x;
    y = copy.y;
    z = copy.z;
    w = copy.w;
}   

// operators
CQuat CQuat::operator* (const CQuat &r) const
{
  return CQuat(
    w*r.x + x*r.w + y*r.z - z*r.y,
    w*r.y + y*r.w + z*r.x - x*r.z,
    w*r.z + z*r.w + x*r.y - y*r.x,
    w*r.w - x*r.x - y*r.y - z*r.z
  );
}

const CQuat & CQuat::operator=( const CQuat & right) {
    x = right.x;
    y = right.y;
    z = right.z;
    w = right.w;

    return *this;
}

/* Construct rotation matrix from (possibly non-unit) quaternion.
 * Assumes matrix is used to multiply column vector on the left:
 * vnew = mat vold.  Works correctly for right-handed coordinate system
 * and right-handed rotations. */
void CQuat::QuatToMatrix(HMatrix& out)
{
    float n = VectorNorm();

    CQuat
        s = (*this) * (n ? (2.f / n) : 0),
        x1 = s * x,
        y1 = s * y,
        z1 = s * z,
        w1 = s * w;

    float a[4][4] = {
        { 1-(y1.y+z1.z),    x1.y-w1.z ,    x1.z+w1.y , 0 },
        {    x1.y+w1.z , 1-(x1.x+z1.z),    y1.z-w1.x , 0 },
        {    x1.z-w1.y ,    y1.z+w1.x , 1-(x1.x+y1.y), 0 },
        {          0 ,          0 ,          0 , 1 } };

    memcpy(out, a, sizeof(HMatrix));
}

/* Return unit magnitude vector in direction of v */
CQuat CQuat::UnitVector()
{
    CQuat u(0.0f, 0.0f, 0.0f);
    float vlen = float(sqrt(VectorNorm()));
    if (vlen != 0.0f) {
	    u.x = x/vlen;
        u.y = y/vlen;
        u.z = z/vlen;
    }
    return u;
}

/* Halve arc between unit vectors v0 and this. */
CQuat CQuat::Bisect(const CQuat & v0)
{
    assert(!w && !v0.w);

    CQuat v = (*this) + v0;
    float n = v.VectorNorm();

    if (n < 1.0e-5)
        return CQuat(0, 0, 1);
    else
        return v * (1.f / ((float)sqrt(n)));
}

/* Return cross product, this x v1 */
CQuat CQuat::CrossProd(const CQuat & v1)
{
    CQuat v;
    v.x = y*v1.z - z*v1.y;
    v.y = z*v1.x - x*v1.z;
    v.z = x*v1.y - y*v1.x;
    v.w = 0.0f;
    return (v);
}

/* Convert window coordinates to sphere coordinates. This CQuat
   represents the mouse. */
CQuat CQuat::MouseOnSphere(const CQuat & ballCenter, double ballRadius)
{
    // ball mouse
    CQuat bm (
        (float)((x - ballCenter.x) / ballRadius),
        (float)((y - ballCenter.y) / ballRadius),
        0);

    double mag = bm.VectorNorm();
    if (mag > 1.0)
    {
        bm *= 1.f / ((float)sqrt(mag));
    }
    else
    {
        bm.z = ((float)sqrt(1 - mag));
    }

    return bm;
}

/* Construct a unit quaternion from two points on unit sphere.  This
   CQuat represents the from point. */
CQuat CQuat::QuatFromBallPoints(const CQuat & to)
{
    return CQuat(
    y*to.z - z*to.y,
    z*to.x - x*to.z,
    x*to.y - y*to.x,
    x*to.x + y*to.y + z*to.z );
}

/* Convert a unit quaternion to two points on unit sphere.
   Returns the arcFrom point. */
void CQuat::QuatToBallPoints(CQuat &arcFrom, CQuat &arcTo)
{
    float s = ((float)sqrt(x*x + y*y));
    arcFrom = s ? CQuat(-y/s, x/s, 0) : CQuat(0, 1, 0);

    arcTo = CQuat(
        w * arcFrom.x - z * arcFrom.y,
        w * arcFrom.y + z * arcFrom.x,
        x * arcFrom.y - y * arcFrom.x );

    if(w < 0)
    {
        arcFrom = CQuat(-arcFrom.x, -arcFrom.y, 0);
    }
}

/* Force sphere point to be perpendicular to axis. This CQuat
   represents the loose CQuat */
CQuat CQuat::ConstrainToAxis(const CQuat & axis)
{
    CQuat onPlane = (*this) - axis * this->DotProd(axis);
    float norm = onPlane.VectorNorm();

    if(norm > 0)
    {
        if(onPlane.z < 0)
        {
            onPlane = -onPlane;
        }

        onPlane *= 1.f / ((float)sqrt(norm));
    }
    else
    {
        onPlane = axis.z==1 ? CQuat(1,0,0) : CQuat(-axis.y,axis.x,0).UnitVector();
    }

    return onPlane;
}

/* Find the index of nearest arc of axis set. The this CQuat * is "loose." */
int CQuat::NearestConstraintAxis(const CQuat *axes, int nAxes)
{
    float max = -1.0f;
    int   nearest = 0;

    for (int i=0; i<nAxes; ++i)
    {
        CQuat onPlane = ConstrainToAxis(axes[i]);
        float dot = onPlane.DotProd(*this);
        if(dot > max)
        {
            max = dot;
            nearest = i;
        }
    }

    return nearest;
}


/////////////////////////////////////////////////////////////////////////////
//  CArcBall Definitions

CQuat CArcBall::Convert(int x, int y, int windowWidth, int windowHeight)
{
    float
    scaleX = (windowWidth>windowHeight) ? ((float)windowWidth/windowHeight) : 1,
    scaleY = (windowHeight>windowWidth) ? ((float)windowHeight/windowWidth) : 1;

    return CQuat(
        ((float)x/windowWidth -.5f)*2.0,// * 2*scaleX,
        -((float)y/windowHeight-.5f)*2.0,// * 2*scaleY,
        0 );
}

// create object
CArcBall::CArcBall () :
    center     (CQuat::ONE),
    radius     (1.0),
    vDown      (CQuat::ONE),
    vNow       (CQuat::ONE),
    qDown      (CQuat::ONE),
    qNow       (CQuat::ONE),
    showResult       (false),
    dragging   (false),
    axisSet    (NoAxes)
{
    memcpy(mNow,  mID, sizeof(HMatrix));
    memcpy(mDown, mID, sizeof(HMatrix));

    sets[CameraAxes] = mID[0];
    setSizes[CameraAxes] = 3;

    sets[BodyAxes] = mDown[0];
    setSizes[BodyAxes] = 3;

    sets[OtherAxes] = OTHER_AXIS[0];
    setSizes[OtherAxes] = 1;

    matrix[0] = 1.0f;  matrix[1] = 0.0f;  matrix[2] = 0.0f;  matrix[3] = 0.0f;
    matrix[4] = 0.0f;  matrix[5] = 1.0f;  matrix[6] = 0.0f;  matrix[7] = 0.0f;
    matrix[8] = 0.0f;  matrix[9] = 0.0f;  matrix[10] = 1.0f;  matrix[11] = 0.0f;
    matrix[12] = 0.0f;  matrix[13] = 0.0f;  matrix[14] = 0.0f;  matrix[15] = 1.0f;
}

// set the center and size of the controller
void CArcBall::BallPlace(const CQuat &c, double r) 
{
  center = c;
  radius = r;
}

// incorporate new mouse position
void CArcBall::BallMouse(const CQuat & vNow1) 
{
    vNow = vNow1;
}

// choose a constraint set, or none
void CArcBall::BallUseSet(AxisSet aSet) 
{
    if(!dragging) 
    {
        axisSet = aSet;
    }
}

// using vDown, vNow, dragging, and axisSet, compute rotations
void CArcBall::BallUpdate()
{
    int setSize = setSizes[axisSet];

    CQuat *set = (CQuat *)(sets[axisSet]);

    vFrom = vDown.MouseOnSphere(center, radius);
    vTo   = vNow.MouseOnSphere(center, radius);

    if(dragging)
    {
        if(axisSet != NoAxes)
        {
            vFrom = vFrom.ConstrainToAxis(set[axisIndex]);
            vTo   = vTo.ConstrainToAxis(set[axisIndex]);
        }
        qDrag = vFrom.QuatFromBallPoints(vTo);
        qNow = qDrag * qDown;
    }
    else if(axisSet != NoAxes) 
    {
        axisIndex = vTo.NearestConstraintAxis(set, setSize);
    }

    qDown.QuatToBallPoints(vrFrom, vrTo);
    CQuat sa = qNow.QuatConj();
    qNow.QuatConj().QuatToMatrix(mNow); // gives transpose to OpenGL
}

// put rotation matrix defined by controller use into our display matrix
void CArcBall::BallValue()
{ 
    memcpy(matrix, mNow, sizeof(HMatrix)); 
}

// drag sequence
void CArcBall::BallBeginDrag()
{
    dragging = true;
    vDown = vNow;
}

void CArcBall::BallEndDrag() 
{
    dragging = false;
    qDown = qNow;

    memcpy(mDown, mNow, sizeof(HMatrix));
}

// draw the controller with all its arcs
void CArcBall::BallDraw(float aspect)
{
    if(dragging) 
    {
        BallDrawOuterRing();  // draw edges
    }
 
    BallDrawResultArc();
    BallDrawConstraints();
    BallDrawDragArc();
}

// draw all constraint arcs
void CArcBall::BallDrawConstraints()
{
    if(axisSet == NoAxes)
    {
        return;
    }

    int setSize = setSizes[axisSet];
    float *set = sets[axisSet];

    CQuat axis;
    for(int axisI=0; axisI<setSize; ++axisI)
    {
        if(axisIndex != axisI)
        {
            if(dragging) 
            {
                continue;
            }
            glColor3d(0.75, 0.5, 0.125);  //farColor
        }
        else
        {
            glColor3d(1.0, 1.0, 0.25);  //nearColor
        }
        
        axis = *(CQuat *)&set[4*axisI];

        if(axis.z == 1) 
        {
            Circ();
        }
        else
        {
            DrawHalfArc(axis);
        }
    }
}

void CArcBall::BallDrawOuterRing()
{
    glColor3d(1.0, 1.0, 1.0);  //rimColor
    Circ();
}

// draw "rubber band" arc during dragging
void CArcBall::BallDrawDragArc()
{
    glColor3d(0.5, 1.0, 1.0);  //dragColor
    if(dragging)
    {
        DrawAnyArc(vFrom, vTo);   
    }
}

// draw arc for result of all drags
void CArcBall::BallDrawResultArc()
{
    glColor3d(0.75, 0.125, 0.125);  //resColor
    if(showResult) 
    {
        DrawAnyArc(vrFrom, vrTo);
    }
}

// draw circle?
void CArcBall::Circ() 
{
    CQuat p(1.0,0,0,0), m(0,1.0,0,0);
    DrawAnyArc( p,  m);
    DrawAnyArc( m, -p);
    DrawAnyArc(-p, -m);
    DrawAnyArc(-m,  p);
}

// draw an arc defined by its ends
void CArcBall::DrawAnyArc(const CQuat &vFrom, const CQuat &vTo)
{
    CQuat pts[NSEGS+1];

    pts[0] = vFrom;
    pts[1] = pts[NSEGS] = vTo;

    double magFrom = sqrt(vFrom.VectorNorm());
    pts[0] /= magFrom;

    for(int i=0; i < LG_NSEGS; ++i)
    {
        pts[1] = pts[0].Bisect(pts[1]);
    }

    float dot = 2 * pts[0].DotProd(pts[1]);
    for(i=2; i < NSEGS; ++i)
    {
        pts[i] = pts[i-1]*dot - pts[i-2];
    }

    glBegin(GL_LINE_STRIP);
    for(i=0; i<=NSEGS; ++i)
    {
        glVertex3f(pts[i].x*magFrom, pts[i].y*magFrom, pts[i].z*magFrom);
    }
    glEnd();
}

// draw the arc of a semi-circle defined by its axis
void CArcBall::DrawHalfArc(const CQuat &n) 
{
    CQuat p, m;

    if(n.z != 1.f)
    {
        p.x=n.y;
        p.y=-n.x;
        p=p.UnitVector();
    }
    else
    {
        p.x=0;
        p.y=1;
    }

    m = p.CrossProd(n);
    DrawAnyArc(p,  m);
    DrawAnyArc(m, -p);
}



