/******************************************************************************
 *
 * ArcBall.h
 *
 * Adapted from starter code used for a Princeton CS assignment, see
 * http://www.cs.princeton.edu/courses/archive/fall01/cs426/assignments/assn3/assn3.html
 *
 * This arcball code is an implementation of the original arcball paper by Ken Shoemake.
 *
 * What the code does is allows 3D rotations (yes all 3 dimensions)
 * through one click of the mouse (in my implementation it's the left
 * mouse button).  This rotation doesn't suffer from Gimbal lock and
 * is really nice to have in general.  In case you want to understand
 * how it interfaces with the Win32 code, see my CBixelViewView class,
 * where the left mouse button clicks interface with the code in 
 * ArcBall.cpp.  Also, understand that the only arcball object I have
 * is a part of CAspectView, which is a member of CBixelViewDoc (if
 * you're not clear on the whole Document/View style of Win32 programming,
 * I would suggest briefly looking over the creating the view and creating
 * the document sections in the MFC "Scribble" Tutorial in the MSDN help 
 * files.  
 *
 * NOTE: I don't fully understand all of this code so I would suggest 
 * looking at other places on the web for better commented ArcBall help,
 * or look at the original paper.  If you don't understand what changes
 * you want to make may do to the program, I strongly suggest that you 
 * back up the code first to avoid the problems that I encountered when 
 * tweaking things in the original code.
 *
 *  02/09/02: D. Hazen--File created 
 *  02/20/02: D. Hazen--Fixed display and rotation problems 
 *  03/19/02: D. Hazen--Finished adding comments to the file
 *
 * Note: Other than comments, the code here is a nearly exact copy of the
 * Princeton code, although the code here has been placed inside a class
 * for a cleaner representation.
 *
 *****************************************************************************/

#ifndef ARCBALL_H
#define ARCBALL_H

#include <assert.h>

// we approximate curved surfaces by a bunch of straight line segments
// this number is the log of the number of those segments
#define LG_NSEGS 4

// this number is the number of segments
#define NSEGS (1<<LG_NSEGS)

// just a nice definition to make for a matrix of floats used commonly
// in OpenGL
typedef float HMatrix[4][4];

// Simple Quaternion class used in ArcBall's calculations
// Again I would suggest searching online for Quaternion
// to get a better understanding of this code
class CQuat : public CObject
{
    friend class CArcBall;
    // stupid unit and zero quaternion definitions just because
    // they're used so often
    static const CQuat ZERO, ONE;

    // method definitions
public:
    // create object
    CQuat (float x_=0, float y_=0, float z_=0, float w_=0) : x(x_), y(y_), z(z_), w(w_) {}
    CQuat(const CQuat & c);

    // simple arithmetic operators for quaternions
    bool  operator== (const CQuat &r) const { return x==r.x && y==r.y && z==r.z && w==r.w; }
    CQuat  operator+  (const CQuat &r) const { return CQuat(x+r.x, y+r.y, z+r.z, w+r.w); }
    CQuat& operator+= (const CQuat &r)       { x+=r.x; y+=r.y; z+=r.z; w+=r.w; return *this; }
    CQuat  operator-  () const              { return CQuat(-x, -y, -z, -w); }
    CQuat  operator-  (const CQuat &r) const { return CQuat(x-r.x, y-r.y, z-r.z, w-r.w); }
    CQuat& operator-= (const CQuat &r)       { x-=r.x; y-=r.y; z-=r.z; w-=r.w; return *this; }
    CQuat  operator*  (float s) const       { return CQuat(x*s, y*s, z*s, w*s); }
    CQuat& operator*= (float s)             { x*=s; y*=s; z*=s; w*=s; return *this; }
    CQuat  operator/  (float s) const       { assert(s); return (*this)* (1/s); }
    CQuat& operator/= (float s)             { assert(s); return (*this)*=(1/s); }
    CQuat  operator*  (const CQuat &r) const;
    CQuat& operator*= (const CQuat &r);
    CQuat  operator/  (const CQuat &r) const { return (*this) *  r.QuatConj() /  r.VectorNorm(); }
    CQuat& operator/= (const CQuat &r)       { return (*this) *= r.QuatConj() /= r.VectorNorm(); }
    const CQuat & operator=( const CQuat &);

    // useful quaternion/vector operations (some use 
    // the idea that a quaternion is just a 4 element vector)
    float DotProd(const CQuat & v1) {return x*v1.x + y*v1.y + z*v1.z + w*v1.w;}
    CQuat CrossProd(const CQuat & v1);
    CQuat Bisect(const CQuat & v0);
    CQuat UnitVector();

    // Normalized Vector
    float VectorNorm() const { return x*x + y*y + z*z + w*w; }

    // Conjugate
    CQuat QuatConj () const { return CQuat(-x, -y, -z, w); }
   
    void QuatToMatrix(HMatrix& out);

    // Ball math
    CQuat MouseOnSphere(const CQuat & ballCenter, double ballRadius);
    CQuat QuatFromBallPoints(const CQuat & to);
    void QuatToBallPoints(CQuat &arcFrom, CQuat &arcTo);
    CQuat ConstrainToAxis(const CQuat & axis);
    int NearestConstraintAxis(const CQuat *axes, int nAxes);

    // member definitions
protected:
    float x;
    float y;
    float z;
    float w;
};

// I think this changes how the camera rotates but for our purposes
// we don't use it (we just set the AxisSet in CArcBall to NoAxes
typedef enum AxisSet{NoAxes, CameraAxes, BodyAxes, OtherAxes, NSets} AxisSet;

// actual arcball class used for the 3D rotations
class CArcBall : public CObject 
{
    
    // method definitions
public:
    CQuat Convert(int x, int y, int windowWidth, int windowHeight);

    // Constructor
    CArcBall();
      
    // set the center and size of the controller
    void BallPlace(const CQuat &c, double r);
    
    // incorporate new mouse position
    void BallMouse(const CQuat & vNow1);

    // choose a constraint set, or none
    void BallUseSet(AxisSet aSet);

    // begin/stop drawing arc for all drags combined
    void BallHideResult() {showResult = false;}
    void BallShowResult() {showResult = true;}

    // using vDown, vNow, dragging, and axisSet, compute rotations
    void BallUpdate();
    
    // return rotation matrix defined by controller use
    void BallValue();

    // drag sequence
    void BallBeginDrag();
    void BallEndDrag();

    // draw the controller with all its arcs
    void BallDraw(float aspect);

    // draw an approximate circle on the screen (a bunch of line 
    // segments based on NSEGS)
    void Circ();

    float * GetMatrix() { return matrix; }
    BOOL GetDragging() { return dragging; }

private:
    // used to draw the arc and circle showing how the user
    // is currently rotating the object (this will make sense
    // if you play around with rotating)
    void DrawAnyArc(const CQuat &vFrom, const CQuat &vTo); 
    void DrawHalfArc(const CQuat &n);
    void BallDrawConstraints();
    void BallDrawOuterRing();
    void BallDrawDragArc();
    void BallDrawResultArc();
    
    // member definitions
protected:
    CQuat center;
    // Radius should stay at 1 for proper results
    double radius;
    CQuat qNow, qDown, qDrag;
    CQuat vNow, vDown, vFrom, vTo, vrFrom, vrTo;
    HMatrix mNow, mDown, mDeltaNow;
    // rotation matrix defined by the arcball
    float matrix[16];
    BOOL showResult, dragging;
    float *(sets[NSets]);
    int setSizes[NSets];
    AxisSet axisSet;
    int axisIndex;
};

#endif